<?
// Class Sequencer
// ===============
// Mode -> behaviour when invoked while a schedule is already being executed
// RESTART:				cancels previous schedule, starts the new one
// TOGGLE:				cancels previous schedule, stops execution
// MULTIPLE:		merges the new schedule with the previous and executes all
// IGNORE_MULTIPLE: as long as a previous schedule is active, ignores new invokations

define('RESTART', 1);
define('TOGGLE', 2);
define('MULTIPLE', 3);
define('IGNORE_MULTIPLE', 4);
define('NOW', time()+1);

class Sequencer {
	//---------------------------------------------------------------------------
	function __construct($mode = RESTART) {
		global $IPS_SELF, $IPS_SENDER;

		$this->is_initialized = false;
		$this->mode = $mode;
		$this->schedule = array();
		$this->continue = true;
		$varName = "schedule_$IPS_SELF";
		$varID = @IPS_GetVariableIDByName($varName, $IPS_SELF);
		$this->varID = $varID;
		if ($varID == 0) {
			$varID = IPS_CreateVariable(3);
			IPS_SetParent($varID, $IPS_SELF);
			IPS_SetName($varID, $varName);
			$this->varID = $varID;

		} elseif ($IPS_SENDER == 'TimerEvent') {
			$this->is_initialized = true;
			$this->schedule = unserialize(GetValueString($varID));

		} else {
			if ($mode == RESTART) {
				$this->cancel_schedule(false);

			} elseif ($mode == TOGGLE) {
				$this->cancel_schedule();
				$this->continue = false;

			} elseif ($mode == MULTIPLE) {
				$this->schedule = unserialize(GetValueString($varID));

			} elseif ($mode == IGNORE_MULTIPLE) {
				$this->continue = false;
			}
		}
	} // __construct

	//---------------------------------------------------------------------------
	function initialize($schedule0) {
		if ($this->continue == false) {
			return;
		}

		$now = time();
		$schedule = array();
		$i = 0;
		$first = true;
		$id0 = (time()-mktime(0,0,0))*100;
		while (($i < count($schedule0)) && (count($schedule0[$i]) >= 1)) {
			$sched = &$schedule0[$i];
			$sched['id'] = $id0 + $i;
			$randDelay = (isset($sched['randomDelay'])) ? rand(0, $sched['randomDelay']) : 0;
			if (isset($sched['randomDelay'])) {
				$randDelay = rand(0, $sched['randomDelay']);
			} else {
				$randDelay = 0;
			}
			if (isset($sched['at'])) {
				$trigTime = $sched['at'] + $randDelay;
				$eid = $this->create_trigger($trigTime);
				$sched['eid'] = $eid;
				$first = false;

			} elseif (isset($sched['at+'])) {
				$trigTime = $now + $sched['at+'] + $randDelay;
				$eid = $this->create_trigger($trigTime);
				$sched['eid'] = $eid;
				$first = false;

			} elseif ($first) {
				if (isset($sched['delay'])) {
					$delay = $sched['delay'] + $randDelay;
					$delay = $this->handle_short_delays($delay);
					$trigTime = $now + min(1,$delay);
					$eid = $this->create_trigger($trigTime);
					$sched['eid'] = $eid;

				} else {
					$trigTime = $now + 1;
					$eid = $this->create_trigger($trigTime);
					$sched['eid'] = $eid;
				}
				$first = false;
			}
			$schedule[] = $sched;
			$i++;
		}
		$schedule = array_merge($this->schedule, $schedule);
		$this->schedule = $schedule;
		SetValueString($this->varID, serialize($schedule));
	} // initialize

	//---------------------------------------------------------------------------
	function exec_scheduled_action() {
	   global $IPS_SENDER, $IPS_EVENT;
		if ((!$this->is_initialized) || ($IPS_SENDER != 'TimerEvent')) {
		   return;
		}
		$schedule = $this->schedule;
		$i = 0;
		$schedNr = false;
		while ($i < count($schedule)) {	 // find record corresponding to current event
			$sched = $schedule[$i];
			if (isset($sched['eid']) && ($sched['eid'] == $IPS_EVENT)) {
				$schedNr = $i;
				array_splice($schedule, $i, 1);
				$this->schedule = $schedule;
				SetValueString($this->varID, serialize($schedule));
				break;
			}
			$i++;
		}

		while ($schedNr !== false) {
			if (isset($sched['eid'])) {
				IPS_DeleteEvent($sched['eid']);
			}
			unset($parameter);
			if (isset($sched['parameter'])) {
				$parameter = $sched['parameter'];
			} else {
				$parameter = '';
			}
			if (isset($sched['function'])) {
				$function = preg_replace('/\(.*\)/','',$sched['function']);
			 $function($parameter);		   // ### Execute action
			}
			if (isset($sched['code'])) {
				$code = $sched['code'];
				eval($code);		 // ### Execute action
			}
			if (isset($sched['eval'])) {
				$code = $sched['eval'];
				eval($code);		 // ### Execute action
			}
			if (isset($sched['exec'])) {
				$cmd = $sched['exec'];
				if (is_array($parameter)) {
					$parameter = implode(" ", $parameter);
				}
				IPS_Execute($cmd, $parameter, false, false);  // ### Execute action
			}

			$ns = $this->find_entry($sched['id']+1);
			if ($ns === false) {
				break;
			}
			$nextSched = &$schedule[$ns];
			$randDelay = (isset($nextSched['randomDelay'])) ? rand(0, $nextSched['randomDelay']) : 0;
			if (! (isset($nextSched['at']) || isset($nextSched['at+']))) {
													 // neither at nor at+ is set:
				if (isset($nextSched['delay'])) {
					$delay = $nextSched['delay'] + $randDelay;
					$delay = $this->handle_short_delays($delay);
					if ($delay > 0) {
						$trigTime = time() + $delay;
					$eid = $this->create_trigger($trigTime);
					$nextSched['eid'] = $eid;
					   SetValueString($this->varID, serialize($schedule));
					   $schedNr = false;
				} else {
					$schedNr = $ns;
					$sched = $schedule[$schedNr];
					array_splice($schedule, $schedNr, 1);
					$this->schedule = $schedule;
					SetValueString($this->varID, serialize($schedule));
				}

			} else {								// no delay, execute immetiately
				$schedNr = $ns;
				$sched = $schedule[$schedNr];
				array_splice($schedule, $schedNr, 1);
				$this->schedule = $schedule;
				SetValueString($this->varID, serialize($schedule));
			}
			} else {
				break;
			}
		}
		if (count($schedule) <= 0) {
			IPS_DeleteVariable($this->varID);
			$schedule = array();
		}
	} // exec_scheduled_action

	//---------------------------------------------------------------------------
	function cancel_schedule($deleteVar = true) {
		global $IPS_SELF;
		$allEvents = IPS_GetEventListByType(1);
		foreach ($allEvents as $e) {
			$eObj = IPS_GetObject($e);
			if ($eObj['ParentID'] == $IPS_SELF) {
				IPS_DeleteEvent($e);
			}
		}
		if ($deleteVar && IPS_VariableExists($this->varID)) {
			IPS_DeleteVariable($this->varID);
		}

	} // cancel_schedule

	//---------------------------------------------------------------------------
	private function create_trigger($time) {
		global $IPS_SELF;
		$eid = IPS_CreateEvent(1);

		IPS_SetEventCyclic($eid, 1, 0, 0, 0, 0, 0);	  // einmaliger Event
		IPS_SetEventCyclicDateBounds($eid, $time, $time );
		IPS_SetEventCyclicTimeBounds($eid, $time, $time );

		IPS_SetParent($eid, $IPS_SELF);
		IPS_SetEventActive($eid, true);
		return $eid;
	} // create_trigger

	//---------------------------------------------------------------------------
	private function handle_short_delays($delay) {
		$delay_fract = intval(($delay - intval($delay))*1000);
		if ($delay_fract > 0) {
			if ($delay < 2) {
				IPS_Sleep(intval($delay*1000));
				$delay = 0;
			} else {
				$delay = intval($delay);
				IPS_Sleep($delay_fract);
			}
		}
		return $delay;
	} // handle_short_delays

	//---------------------------------------------------------------------------
	private function find_entry($id) {
		$schedule = $this->schedule;
		$i = 0;
		while ($i < sizeof($schedule)) {
			$sched = &$schedule[$i];
			if (isset($sched['id']) && ($sched['id'] == $id)) {
				return $i;
			}
			$i++;
		}
		return false;
	} // find_entry
} // class Sequencer

//---------------------------------------------------------------------------
function cancel_schedule($scriptID) {
	$allEvents = IPS_GetEventListByType(1);
	foreach ($allEvents as $e) {
		$eObj = IPS_GetObject($e);
		if ($eObj['ParentID'] == $scriptID) {
			IPS_DeleteEvent($e);
		}
	}
	$varName = "schedule_$scriptID";
	$varID = @IPS_GetVariableIDByName($varName, $scriptID);
	if ($varID != 0) {
		IPS_DeleteVariable($varID);
	}
} // cancel_schedule


//---------------------------------------------------------------------------
function show_schedule($schedule) {
	$str = "Schedule:\n";
	foreach ($schedule as $sched) {
		$str .= "  ".var_r($sched, true)."\n";
	}
	return $str;
} // show_schedule

?>